# ==========================
# Module: Regression Tasks
# Author: Kenneth Leung
# Last Modified: 22 Feb 2022
# ==========================
import pandas as pd
import statsmodels.api as sm
# from statsmodels.genmod.generalized_linear_model import GLM
# from statsmodels.genmod import families


def task_linear_regression(y: pd.Series,
                           X: pd.DataFrame):
    """Run linear regression task (ordinary least squares) with statsmodel library

    Args:
        y (pd.Series): Series of target variable
        X (pd.DataFrame): DataFrame of predictor variables

    Returns:
        Returns the regression model residuals, fitted values, and
        the OLS regression summary table
    """

    model = sm.OLS(y, X)
    results = model.fit()
    residuals = results.resid
    fitted = results.fittedvalues
    summary = results.summary()

    return residuals, fitted, summary


# # Binary Logistic Regression Assumption Checks - COMING SOON
# def task_binary_logistic_regression(y: pd.Series,
#                                     X: pd.DataFrame,
#                                     task_summary: bool = True):
#     model = GLM(y, X, family=families.Binomial())
#     results = model.fit()
#     summary = results.summary()

#     return results, summary


# # Multinomial Logistic Regression Assumption Checks - COMING SOON
# def task_multinomial_logistic_regression(y: pd.Series,
#                                          X: pd.DataFrame):
#     # model = GLM(y, X, family=families.Binomial()) # To amend to MN logit
#     print('Multinomial Logistic Regression Assumption Checks - COMING SOON')



# ==========================
# Module: Diagnostic Plots
# Author: Kenneth Leung
# Last Modified: 06 Jan 2022
# ==========================
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import statsmodels.api as sm
import statsmodels.tsa.api as smt
import warnings

from .utils import display_base64_plot
warnings.filterwarnings("ignore")


# ---------------------
#   Homoscedasticity
# ---------------------
def plot_residual_fit(fitted: pd.Series,
                      residuals: pd.Series,
                      check_type: str,
                      batch_size: int = 10):
    """Generate and display the regression residual plot (fitted vs residuals)

    Args:
        fitted (pd.Series): Fitted (aka predicted) values from model
        residuals (pd.Series): Residual values from model
        check_type (str): Set the residual plot layout specific to the assumption check (i.e. homoscedasticity or linearity)
        batch_size (int, optional): Number of data points used to compute each quantile point. Defaults to 10.

    Returns:
        base64 object: Encoded image of residual plot
    """

    sns.set_style('white')  # Define seaborn layout theme
    upper_quantile_value, lower_quantile_value = 80, 20  # Set quantile ranges for upper and lower quantile lines

    fig, ax = plt.subplots(1, 1, figsize=(6, 6))
    ax.set(xlabel='Fitted', ylabel='Residual')

    # Scatter plot of fitted (x axis) vs residuals (y axis)
    plt.scatter(fitted,
                residuals,
                color='gray',
                alpha=0.6)

    # Show horizontal dotted line at y=0
    plt.axhline(y=0, color='black', linestyle='dotted')

    # Display different types of lines based on whether it is homoscedasticity or linearity check
    if check_type == 'linearity':
        reg_line, outer_quantiles = True, False
    elif check_type == 'homosced':
        reg_line, outer_quantiles = False, True
    else:
        reg_line, outer_quantiles = True, True

    if reg_line:
        sns.regplot(x=fitted,
                    y=residuals,
                    lowess=True,
                    ax=ax,
                    scatter=False,
                    line_kws={'color': 'red'})

    if outer_quantiles:
        resid_fit_df = pd.concat([fitted, residuals], axis=1)
        resid_fit_df.columns = ['fitted', 'residuals']
        resid_fit_df.sort_values(by='fitted', inplace=True)
        resid_fit_df.reset_index(drop=True, inplace=True)

        quantile_df = pd.DataFrame()
        indices = resid_fit_df.index.tolist()

        # Generate data points to plot quantile lines
        for i in range(0, len(indices), batch_size):
            index_array = indices[i:i + batch_size]
            mid_index = index_array[len(index_array) // 2]
            fitted_mid_value = resid_fit_df['fitted'].iloc[mid_index]
            resid_mid_value = resid_fit_df['residuals'].iloc[mid_index]
            resid_array = resid_fit_df['residuals'].iloc[index_array].values.tolist()
            upper_quantile = np.percentile(resid_array, upper_quantile_value)
            lower_quantile = np.percentile(resid_array, lower_quantile_value)

            resid_dict = {'fitted': fitted_mid_value,
                          'residual': resid_mid_value,
                          'lower_quantile': lower_quantile,
                          'upper_quantile': upper_quantile}

            quantile_df = quantile_df.append(resid_dict, ignore_index=True)

        sns.regplot(x=quantile_df['fitted'].values,
                    y=quantile_df['upper_quantile'].values,
                    lowess=True,
                    ax=ax,
                    scatter=False,
                    line_kws={'color': '#0171C0'})

        sns.regplot(x=quantile_df['fitted'].values,
                    y=quantile_df['lower_quantile'].values,
                    lowess=True,
                    ax=ax,
                    scatter=False,
                    line_kws={'color': '#0171C0'})

    # Save plot as base64 image object for display
    output_plot = display_base64_plot(fig)

    return output_plot


# ---------------------
#    Independence
# ---------------------
def plot_acf(residuals: pd.Series,
             alpha: float = 0.05):
    """Generate and display Auto-Correlation Function (ACF) plot

    Args:
        residuals (pd.Series): Residual values from model
        alpha (float, optional): The level for obtaining the confidence intervals.
            For e.g.,if alpha=0.05, 95% confidence intervals are returned.
            Defaults to 0.05.

    Returns:
        base64 object: Encoded image of ACF plot
    """

    fig, ax = plt.subplots(1, 1, figsize=(6, 6))
    smt.graphics.plot_acf(residuals, alpha=alpha, ax=ax)

    # Save plot as base64 image object for display
    output_plot = display_base64_plot(fig)

    return output_plot


# ---------------------
#      Linearity
# ---------------------
def plot_pairplot(df: pd.DataFrame,  # Main dataframe
                  target: str,
                  hide_categorical: bool = True,
                  reg_line: bool = False):
    """Generate and display pairplot for visualization of pair-wise linear relationships

    Args:
        df (pd.DataFrame): Dataframe of pre-processed data containing predictors and target variables
        hide_categorical (bool, optional): Hide the categorical variables from being displayed
            in pairplots. Defaults to True.
        reg_line (bool, optional): Whether to include regression lines in the pairplots. Defaults to False.

    Returns:
        base64 object: Encoded image of pair-plots
    """

    if hide_categorical:
        X_cols = [col for col in df.columns.tolist() if col != target]
        bool_cols = [col for col in df if np.isin(df[col].dropna().unique(), [0, 1]).all()]
        X_cols = [item for item in X_cols if item not in bool_cols]

    if reg_line:
        plot_kind = 'reg'  # Show regression line
    else:
        plot_kind = 'scatter'  # Show scatter points without regression line

    pairplot = sns.pairplot(data=df,
                            y_vars=target,
                            x_vars=X_cols,
                            aspect=0.9,
                            kind=plot_kind)

    fig = pairplot.fig

    # Save plot as base64 image object for display
    output_plot = display_base64_plot(fig)

    return output_plot


# ---------------------
#   Multicollinearity
# ---------------------
def plot_corr_heatmap(X: pd.DataFrame,
                      corner: bool,
                      cmap: str = "YlGnBu",
                      axes_style: str = 'white'):
    """Generate and display correlation heatmap to assess correlation amongst variables

    Args:
        X (pd.DataFrame): Dataframe containing only predictor variables (i.e. exclude target variable)
        corner (bool): Display diagonal heatmap (as opposed to standard square heatmap)
        cmap (str, optional): Colour map of correlation heatmap. Defaults to "YlGnBu".
        axes_style (str, optional): Layout style of heatmap plot axes. Defaults to 'white'.

    Returns:
        base64 object: Encoded image of ACF plot
    """

    # Generate correlation values
    corrMatrix = X.corr()
    fig, ax = plt.subplots(figsize=(7, 7))
    if corner is False:
        sns.heatmap(corrMatrix,
                    annot=True,
                    square=True,
                    cmap=cmap)
    else:
        mask = np.zeros_like(corrMatrix)
        mask[np.triu_indices_from(mask)] = True

        with sns.axes_style(axes_style):
            sns.heatmap(corrMatrix,
                        mask=mask,
                        annot=True,
                        vmax=.3,
                        square=True,
                        cmap=cmap)

    # Save plot as base64 image object for display
    output_plot = display_base64_plot(fig)

    return output_plot


# ---------------------
#      Normality
# ---------------------
def plot_qq(residuals: pd.Series):
    """Generate and display Quantile-Quantile (QQ) plot

    Args:
        residuals (pd.Series): Residual values from model

    Returns:
        base64 object: Encoded image of QQ plot
    """

    fig, ax = plt.subplots(1, 1, figsize=(6, 6))
    fig = sm.ProbPlot(residuals).qqplot(line='s',
                                        ax=ax,
                                        marker='.',
                                        markersize='11',
                                        markerfacecolor='gray',
                                        markeredgecolor='gray',
                                        alpha=0.8)

    # Save plot as base64 image object for display
    output_plot = display_base64_plot(fig)

    return output_plot


def plot_residual_histogram(residuals: pd.Series):
    """Generate and display histogram of residuals

    Args:
        residuals (pd.Series): Residual values from model

    Returns:
        base64 object: Encoded image of residual histogram
    """
    fig, ax = plt.subplots(1, 1, figsize=(5, 5))

    sns.histplot(residuals,
                 kde=True,
                 ax=ax)

    # Save plot as base64 image object for display
    output_plot = display_base64_plot(fig)

    return output_plot


# ---------------------
#      Outliers
# ---------------------
# def check_outliers(results,
#                    residuals):

#     # Get influence measures
#     influence = results.get_influence()

#     # Obtain summary df of influence measures
#     summary_df = influence.summary_frame()

#     # # Filter summary df to Cook distance
#     cook_df = summary_df.loc[:,['cooks_d']]

#     # Append absolute standardized residual values
#     cook_df['std_resid'] = stats.zscore(results.resid_pearson)
#     cook_df['std_resid'] = cook_df.loc[:,'std_resid'].apply(lambda x: np.abs(x))

#     # Sort by Cook's Distance
#     cook_df.sort_values("cooks_d", ascending=False, inplace=True)

#     # Set Cook's distance threshold
#     cook_threshold = round(4 / len(residuals),3)

#     print(f"Threshold for Cook's Distance = {cook_threshold}")
#     print(f"Displaying Top 5 observations based on Cook's Distance")

#     # Plot influence measures (Cook's distance)
#     fig = influence.plot_index(y_var="cooks", threshold=cook_threshold)
#     fig.tight_layout(pad=1)
#     plt.axhline(y=cook_threshold, ls="--", color='red')
#     plt.show()



# ===============================
# Module: Dashboard Report Generation
# Author: Kenneth Leung
# Last Modified: 22 Feb 2022
# ===============================
import sys
from dataclasses import dataclass
from typing import List, Optional
import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.utils.multiclass import type_of_target
from sklearn.preprocessing import OrdinalEncoder
from jupyter_dash import JupyterDash
from dash.dependencies import Input, Output
import dash_bootstrap_components as dbc
import warnings
from .layouts.linear_regression.layout import layout_linear_regression
from .layouts.linear_regression.tabs import *
from .layouts.placeholder.layout import layout_placeholder
from .layouts.placeholder.tabs import *
from .enums import TaskType
from .plots import *
from .stats import *
from .tasks import *
from .logger import logger

sys.path.insert(0, "/layouts")
warnings.filterwarnings("ignore")


@dataclass
class Check:
    df: pd.DataFrame
    target: str
    task: Optional[str] = None
    predictors: Optional[List[str]] = None
    keep: Optional[bool] = True
    categorical_features: Optional[List[str]] = None
    categorical_encoder: Optional[str] = None
    mode: str = 'inline'  # Other options: external, jupyterlab

    """Core object of StatsAssume. Initializes regression modelling, runs relevant assumption checks,
    and returns an output report in the form of a Dash dashboard
    """
    def __post_init__(self):
        pass
        # if os.path.exists(self.output):
        #     raise Exception("Output directory already exists. Please specify another folder name.")
        # os.makedirs(self.output, exist_ok=True)
        # logger.info(f"Output directory: {self.output}")
        # To include output: Optional[str] = 'output' inside parameters if using this function

    def _determine_task_type(self):
        """Sets type of regression task to run.
        If none specified, function will automatically
        infer type of regression based on target variable

        Raises:
            ValueError: If input string is not a valid regression task
            Exception: If target variable is not in the right format (i.e. 1D-array or column vector)
            Exception: If unable to infer type of regression based on target variable

        Returns:
            Type of regression modelling task to be executed
        """
        if self.task is not None:
            if self.task == 'linear regression':
                task_type = TaskType.linear_regression
            elif self.task == 'binary logistic regression':
                task_type = TaskType.binary_logistic_regression
            elif self.task == 'multinomial logistic regression':
                task_type = TaskType.multinomial_logistic_regression
            else:
                raise ValueError('Unable to infer type of regression task. If specifying manually, please choose from the following: \
                    "linear regression", "binary logistic regression" or "multinomial logistic regression"')

            logger.info(f"[+] Executing task type (specified by user): {task_type.name}")

        else:
            # Infer target type
            target_type = type_of_target(self.df[self.target].values)  # Using sklearn.utils.multiclass.type_of_target
            if target_type == "continuous":
                task_type = TaskType.linear_regression
            elif target_type == "binary":
                task_type = TaskType.binary_logistic_regression
            elif target_type == "multiclass":
                task_type = TaskType.multinomial_logistic_regression
            elif (target_type == "continuous-multioutput") or (target_type == "multilabel-indicator"):
                raise Exception('Target needs to be a 1D-array or column vector')
            else:
                raise Exception("Target type not supported. Please provide a Target variable with type compatible with linear or logistic regression")

            logger.info(f"[+] Executing task type (detected automatically): {task_type.name}")

        return task_type

    def _keep_predictors(self):
        """ Keeps or drops variables based on list of variables specified by user

        Returns:
            pd.DataFrame: Output dataframe after keeping/dropping predictor variables
        """
        if self.predictors is not None:
            # Keep defined columns
            if self.keep:
                selected_cols = self.predictors + [self.target]
                df_keep_predictors = self.df[selected_cols]
                logger.info(f'[+] Keep predictor variables specified by user: {selected_cols}')
            # Drop defined columns
            else:
                df_keep_predictors = self.df.drop(self.predictors, axis=1)
                logger.info(f'[+] Drop variables specified by user: {self.predictors}')
        else:
            df_keep_predictors = self.df

        return df_keep_predictors

    def _process_categorical_features(self, df):
        """Transform categorical features (via one-hot or ordinal encoding)
        so that input data can be parsed into statsmodel for modelling

        Args:
            df (pd.DataFrame): DataFrame with defined predictor and target variables
            categorical_features (list): List of features that are of categorical type (Manually defined)

        Raises:
            Exception: If categorical variables remain non-encoded

        Returns:
            pd.DataFrame: Output dataframe after encoding of categorical features
        """
        # Automatically detect categorical features
        categorical_features_auto = []
        features = [col for col in list(df.columns) if col != self.target]
        for col in features:
            if self.df[col].dtype == "object":
                categorical_features_auto.append(col)

        # Categorical features specified by user
        if self.categorical_features is not None:
            # Check whether there are additional categorical features not specified by user:
            categorical_features_diff = list(set(categorical_features_auto) - set(self.categorical_features))
            if len(categorical_features_diff) > 0:
                raise ValueError(f'These categorical features have not been encoded: {categorical_features_diff}')
            else:
                logger.info(f'[+] Categorical features specified by user: {categorical_features_auto}')
                categorical_features = self.categorical_features

        # No categorical features specified by user
        else:
            logger.info(f'[+] Categorical features automatically identified: {categorical_features_auto}')
            categorical_features = categorical_features_auto

        if len(categorical_features) > 0:
            if self.categorical_encoder == 'ohe':
                logger.info('[+] Proceeding with one-hot encoding of categorical features')
                df_encode = self.df.copy()
                for feature in categorical_features:
                    dummies = pd.get_dummies(df_encode[feature],
                                             prefix=feature,
                                             drop_first=True)
                    df_encode = pd.concat([df_encode, dummies], axis=1)
                    df_encode.drop([feature], axis=1, inplace=True)

                logger.info(f'[+] Completed one-hot encoding of categorical features: {categorical_features}')
                return df_encode

            elif self.categorical_encoder == 'ord':  # Ordinal encoding
                df_encode = self.df.copy()
                logger.info('[+] Proceeding with ordinal encoding of categorical features')
                ord_encoder = OrdinalEncoder(handle_unknown="use_encoded_value",
                                             unknown_value=np.nan)
                df_encode[categorical_features] = ord_encoder.fit_transform(df_encode[categorical_features].values)
                logger.info(f'[+] Completed ordinal encoding of categorical features: {categorical_features}')

                return df_encode
            else:
                raise ValueError(f'Please manually encode these categorical features (or use `categorical_encoder`) before continuing: {categorical_features} ')
        else:
            logger.info('[+] No non-encoded categorical features identified. Proceeding with regression')
            return self.df

    def report(self):
        """Run regression modeling and assumption checks, and display output report as JupyterDash dashboard

        Raises:
            Exception: If input string of regression type is not recognized

        Returns:
            JupyterDash dashboard (running on localhost server port 8090)
        """

        task_type = self._determine_task_type()
        df = self._keep_predictors()
        df = self._process_categorical_features(df)

        app = JupyterDash(__name__,
                          external_stylesheets=[dbc.themes.BOOTSTRAP]
                          )

        predictors = [col for col in list(df.columns) if col != self.target]
        X, y = df[predictors], df[self.target]
        X_constant = sm.add_constant(X)  # Add constant value for X

        if task_type == TaskType.linear_regression:
            task_name = 'Linear Regression'
            app.layout = layout_linear_regression
            residuals, fitted, summary_str = task_linear_regression(y, X_constant)

            @app.callback(Output('tab-content', 'children'),
                          [Input('tabs', 'value')])
            def render_content(tab):
                if tab == 'tab_summary':
                    return generate_tab_summary(summary_str, task_name)
                elif tab == 'tab_homosced':
                    return generate_tab_homosced(residuals, fitted, X_constant)
                elif tab == 'tab_independence':
                    return generate_tab_independence(residuals)
                elif tab == 'tab_linearity':
                    return generate_tab_linearity(df, self.target, residuals, fitted)
                elif tab == 'tab_multicollinearity':
                    return generate_tab_multicollinearity(X, X_constant)
                elif tab == 'tab_normality':
                    return generate_tab_normality(residuals)

        elif task_type == TaskType.binary_logistic_regression:
            task_name = 'Binary Logistic Regression'
            # task_binary_logistic_regression(y, X_constant)
            app.layout = layout_placeholder

            @app.callback(Output('tab-content', 'children'),
                          [Input('tabs', 'value')])
            def render_content(tab):
                if tab == 'tab_summary':
                    # Dummy placeholder layout
                    return generate_placeholder_layout(task_name)

        elif task_type == TaskType.multinomial_logistic_regression:
            task_name = 'Multinomial Logistic Regression'
            # task_multinomial_logistic_regression(y, X_constant)
            app.layout = layout_placeholder

            @app.callback(Output('tab-content', 'children'),
                          [Input('tabs', 'value')])
            def render_content(tab):
                if tab == 'tab_summary':
                    # Dummy placeholder layout
                    return generate_placeholder_layout(task_name)

        else:
            raise ValueError('Task type not recognized')

        app.run_server(mode=self.mode, port=8090)



# ===============================================
# Module: Dashboard Layout (Placeholder)
# Author: Kenneth Leung
# Last Modified: 13 Feb 2022
# ==============================================
from dash import html, dcc

# Setup HTML layout for dashboard report
layout_placeholder = html.Div([
    html.P(dcc.Tabs(id="tabs", value='tab_summary',
                    children=[dcc.Tab(label='Coming Soon!',
                                      value='tab_summary',
                                      className='tab_summary')
                              ],
                    colors={"border": "white",
                            "primary": "gold",
                            "background": "#F7F7F7"
                            },
                    style={'font-size': '16px',
                           'text-align': 'center',
                           'border-width': '4px',
                           'padding': '0px 0px 0px 0px'
                           }
                    )),
    html.Div(id='tab-content'),
])



# ===============================================
# Module: Dashboard Layout for Linear Regression
# Author: Kenneth Leung
# Last Modified: 02 Jan 2022
# ==============================================
from dash import html, dcc

# Setup HTML layout for linear regression dashboard report
layout_linear_regression = html.Div([
    html.P(dcc.Tabs(id="tabs", value='tab_summary',
                    children=[dcc.Tab(label='Regression Summary',
                                      value='tab_summary',
                                      className='tab_summary'),
                              dcc.Tab(label='Linearity',
                                      value='tab_linearity',
                                      className='tab_linearity'),
                              dcc.Tab(label='Independence',
                                      value='tab_independence',
                                      className='tab_independence'),
                              dcc.Tab(label='Multicollinearity',
                                      value='tab_multicollinearity',
                                      className='tab_multicollinearity'),
                              dcc.Tab(label='Normality',
                                      value='tab_normality',
                                      className='tab_normality'),
                              dcc.Tab(label='Homoscedasticity',
                                      value='tab_homosced',
                                      className='tab_homosced'),
                              ],
                    colors={"border": "white",
                            "primary": "gold",
                            "background": "#F7F7F7"
                            },
                    style={'font-size': '16px',
                           'text-align': 'center',
                           'border-width': '4px',
                           'padding': '0px 0px 0px 0px'
                           }
                    )),
    html.Div(id='tab-content'),
])



# ========================================
# Module: Dash Tabs for Linear Regression
# Author: Kenneth Leung
# Last Modified: 22 Feb 2022
# =======================================
from dash import html
import pandas as pd
from urllib.request import urlopen

from ...plots import *
from ...stats import *
from ...utils import *
from ...explainers import *

# Define regression task
task = 'linear_regression'


# Define function to retrieve example figures
def _get_example_fig(task: str,
                     check: str):
    """Retrieve base64 encoded file of example figures (for visual plot comparison)

    Args:
        task (str): Name of task (e.g. linear_regression)
        check (str): Name of check (e.g. normality)

    Returns:
        base64 encoded file of png image
    """
    fig_url = f'https://raw.githubusercontent.com/kennethleungty/statsassume/main/figures/{task}'

    return base64.b64encode(urlopen(f'{fig_url}/{check}.png').read())

    # Testing
    # fig_url = 'https://raw.githubusercontent.com/kennethleungty/COVID19-Vaccine-Sentiment-Analysis/main/Images/composite_sentiments_by_max_voting.png'
    # return base64.b64encode(urlopen(f'{fig_url}').read())


def generate_tab_summary(summary,
                         task_type: str):
    """Display OLS regression summary as HTML

    Args:
        summary (statsmodels.iolib.summary.Summary): statsmodel OLS regression summary (comprising 3 tables)
        task_type (str): Type of regression task that was executed

    Returns:
        html.Div: OLS regression summary
    """

    # TODO: Need to adjust this to accomodate logistic regression in the future
    ols_table_1 = summary.tables[0].as_html()
    ols_table_2 = summary.tables[1].as_html()
    ols_table_3 = summary.tables[2].as_html()

    tab_summary = html.Div(children=[
                           display_tab_header(assumption_name='Summary',
                                              assumption_intro='Brief description of regression results'),
                           html.Br(),
                           display_regression_summary('Linear Regression Results',
                                                      ols_table_1,
                                                      ols_table_2,
                                                      ols_table_3),
                           html.Br(),
                           display_logs(),
                           html.Br(),
                           display_assumption_details(explain_description_linear_regression,
                                                      explain_solution_linear_regression)
                           ],
                           style={'text-align': 'center'})

    return tab_summary


def generate_tab_homosced(residuals: pd.Series,
                          fitted: pd.Series,
                          X_constant: pd.DataFrame):
    """Generate and display HTML tab report of homoscedasticity assumption check

    Args:
        residuals (pd.Series): Residual values of regression model
        fitted (pd.Series): Fitted (aka predicted) values from model
        X_constant (pd.DataFrame): Predictor dataframe with constant (intercept) included

    Returns:
        html.Div: Report of homoscedasticity assumption check
    """

    check_type = 'homosced'
    output_plot_residual = plot_residual_fit(residuals, fitted, check_type)
    fig_homosced = _get_example_fig(task, check_type)
    interpretation_bp, table_bp = stat_breuschpagan(residuals, X_constant)
    interpretation_white, table_white = stat_white(residuals, X_constant)
    # output_str_gq = stat_gq(residuals, X_constant)

    tab_homosced = html.Div(style={'text-align': 'center'},
                            children=[
                            display_tab_header(assumption_name='Assumption Check - Homoscedasticity',
                                               assumption_intro='Brief description of specific assumption check'),
                            html.Br(),
                            display_visual_plot(plot_name='Residual Plot',
                                                img_plot=output_plot_residual,
                                                img_examples=f'data:image/png;base64,{fig_homosced.decode()}',
                                                explainer=explain_plot_residual_homosced),
                            html.Br(),
                            display_stat_results(test_name='Breusch-Pagan Test',
                                                 interpretation=interpretation_bp,
                                                 test_table=table_bp),
                            html.Br(),
                            display_stat_results(test_name='White Test',
                                                 interpretation=interpretation_white,
                                                 test_table=table_white),
                            html.Br(),
                            display_assumption_details(explain_description_homosced,
                                                       explain_solution_homosced)
                            ])

    return tab_homosced


def generate_tab_independence(residuals: pd.Series):
    """Generate and display HTML tab report of independence assumption check

        Args:
            residuals (pd.Series): Residual values of regression model

        Returns:
            html.Div: Report of independence assumption check
        """

    check_type = 'independence'
    interpretation_dw, table_dw = stat_durbin_watson(residuals)
    interpretation_lb, table_lb = stat_ljungbox(residuals)
    output_plot_acf = plot_acf(residuals)
    fig_independence = _get_example_fig(task, check_type)

    tab_independence = html.Div(style={'text-align': 'center'},
                                children=[
                                display_tab_header(assumption_name='Assumption Check - Independence',
                                                   assumption_intro='Brief description of specific assumption check'),
                                html.Br(),
                                display_visual_plot(plot_name='Autocorrelation Function (ACF) Plot',
                                                    img_plot=output_plot_acf,
                                                    img_examples=f'data:image/png;base64,{fig_independence.decode()}',
                                                    explainer=explain_plot_acf),
                                html.Br(),
                                display_stat_results(test_name='Durbin-Watson Test',
                                                     interpretation=interpretation_dw,
                                                     test_table=table_dw),
                                html.Br(),
                                display_stat_results(test_name='Ljung-Box Test',
                                                     interpretation=interpretation_lb,
                                                     test_table=table_lb),
                                html.Br(),
                                display_assumption_details(explain_description_independence,
                                                           explain_solution_independence)
                                ])

    return tab_independence


def generate_tab_linearity(df: pd.DataFrame,
                           target: str,
                           residuals: pd.Series,
                           fitted: pd.Series):
    """Generate and display HTML tab report of linearity assumption check

    Args:
        df (pd.DataFrame): Dataframe of pre-processed data containing predictors and target variables
        target (str): Name of target variable
        residuals (pd.Series): Residual values from regression model
        fitted (pd.Series): Fitted (aka predicted) values from model

    Returns:
        html.Div: Report of linearity assumption check
    """

    check_type = 'linearity'
    output_plot_residual = plot_residual_fit(residuals, fitted, check_type)
    fig_linearity = _get_example_fig(task, check_type)
    output_plot_pairplot = plot_pairplot(df, target)

    tab_linearity = html.Div(style={'text-align': 'center'},
                             children=[
                             display_tab_header(assumption_name='Assumption Check - Linearity',
                                                assumption_intro='Brief description of specific assumption check'),
                             html.Br(),
                             display_visual_plot(plot_name='Residual Plot',
                                                 img_plot=output_plot_residual,
                                                 img_examples=f'data:image/png;base64,{fig_linearity.decode()}',
                                                 explainer=explain_plot_residual_linearity),
                             html.Br(),
                             display_visual_plot(plot_name='Pair-Plot',
                                                 img_plot=output_plot_pairplot,
                                                 img_examples='None',
                                                 explainer=explain_plot_pairplot),
                             html.Br(),
                             display_assumption_details(explain_description_linearity,
                                                        explain_solution_linearity)
                             ])

    return tab_linearity


def generate_tab_multicollinearity(X: pd.DataFrame,
                                   X_constant: pd.DataFrame):
    """Generate and display HTML tab report of multi-collinearity assumption check

    Args:
        X (pd.DataFrame): Dataframe with predictor variables (without intercept)
        X_constant (pd.DataFrame): Dataframe with predictor variables, and with constant (intercept) included

    Returns:
        html.Div: Report of multi-collinearity assumption check
    """

    output_plot_corr_heatmap = plot_corr_heatmap(X, corner=True)  # Show corr matrix as corner
    interpretation_vif, table_vif = stat_vif(X_constant)

    tab_multicollinearity = html.Div(style={'text-align': 'center'},
                                     children=[
                                     display_tab_header(assumption_name='Assumption Check - Multicollinearity',
                                                        assumption_intro='Brief description of specific assumption check'),
                                     html.Br(),
                                     display_visual_plot(plot_name='Correlation Matrix',
                                                         img_plot=output_plot_corr_heatmap,
                                                         img_examples='None',
                                                         explainer=explain_plot_corrmatrix),
                                     html.Br(),
                                     display_stat_results(test_name='Variance Inflation Factor (VIF)',
                                                          interpretation=interpretation_vif,
                                                          test_table=table_vif),
                                     html.Br(),
                                     display_assumption_details(explain_description_multicollinearity,
                                                                explain_solution_multicollinearity)
                                     ])

    return tab_multicollinearity


def generate_tab_normality(residuals: pd.Series):
    """Generate and display HTML tab report of residual normality assumption check

    Args:
        residuals (pd.Series): Residual values from regression model

    Returns:
        html.Div: Report of residual normality assumption check
    """
    check_type = 'normality'
    output_plot_residual_histogram = plot_residual_histogram(residuals)
    output_plot_qq = plot_qq(residuals)
    fig_normality = _get_example_fig(task, check_type)
    interpretation_shapiro, table_shapiro = stat_shapiro(residuals)

    # Only showing the main statistical test i.e. Shapiro
    # output_str_anderson = stat_anderson(residuals)
    # output_str_jarque_bera = stat_jarque_bera(residuals)
    # output_str_lilliefors = stat_lilliefors(residuals)

    tab_normality = html.Div(style={'text-align': 'center'},
                             children=[
                             display_tab_header(assumption_name='Assumption Check - Normality',
                                                assumption_intro='Brief description of specific assumption check'
                                                ),
                             html.Br(),
                             display_visual_plot(plot_name='Quantile-Quantile (QQ) Plot',
                                                 img_plot=output_plot_qq,
                                                 img_examples=f'data:image/png;base64,{fig_normality.decode()}',
                                                 explainer=explain_plot_qq
                                                 ),
                             html.Br(),
                             display_visual_plot(plot_name='Histogram of Residuals',
                                                 img_plot=output_plot_residual_histogram,
                                                 img_examples='None',
                                                 explainer=explainer_plot_residual_histogram
                                                 ),
                             html.Br(),
                             display_stat_results(test_name='Shapiro-Wilk Test',
                                                  interpretation=interpretation_shapiro,
                                                  test_table=table_shapiro
                                                  ),
                             html.Br(),
                             display_assumption_details(explain_description_normality,
                                                        explain_solution_normality
                                                        )
                             ])

    return tab_normality



# ==========================
# Module: Datasets
# Author: Kenneth Leung
# Last Modified: 12 Jan 2022
# ==========================
import pandas as pd


def load_data(dataset_name: str,
              processed: bool = False,
              save_copy: bool = False,
              raw_url: str = 'https://raw.githubusercontent.com/kennethleungty/statsassume/main/datasets/',
              file_ext: str = '.csv'):
    """Loads toy dataset for assumption checks

    Args:
        dataset_name (str): Name of dataset (selected from list of available datasets)
        processed (bool, optional): If True, retrieves the processed data version instead of raw one. Defaults to False.
        save_copy (bool, optional): Save a copy of dataset locally. Defaults to False.
        raw_url (str, optional): URL where datasets are stored. Defaults to 'https://raw.githubusercontent.com/kennethleungty/Logistic-Regression-Assumptions/main/datasets/'.
        file_ext (str, optional): Extension of data file. Defaults to '.csv'.

    Returns:
        pd.DataFrame: Dataframe of the retrieved toy dataset
    """

    if processed:
        try:
            filename = dataset_name + '_processed'
            data = pd.read_csv(raw_url + filename + file_ext)
        except Exception:
            pass
        else:
            filename = dataset_name
            data = pd.read_csv(raw_url + dataset_name + file_ext)
    else:
        filename = dataset_name
        data = pd.read_csv(raw_url + filename + file_ext)

    if save_copy:
        data.to_csv(filename, index=False)

    return data



# ==========================
# Module: Enums - Task Type
# Author: Kenneth Leung
# Last Modified: 02 Jan 2022
# ==========================
import enum


class TaskType(enum.IntEnum):
    linear_regression = 1
    binary_logistic_regression = 2
    multinomial_logistic_regression = 3

    @staticmethod
    def from_str(task_type: str):
        if task_type == 'linear regression':
            return TaskType.linear_regression
        elif task_type == 'binary logistic regression':
            return TaskType.binary_logistic_regression
        elif task_type == 'multinomial logistic regression':
            return TaskType.multinomial_logistic_regression
        else:
            raise ValueError(f'Specified task type of {task_type} is invalid')

    @staticmethod
    def list_str():
        return ['linear regression', 'binary logistic regression', 'multinomial logistic regression']



# ==========================
# Module: Logging
# Author: Kenneth Leung
# Last Modified: 30 Jan 2022
# ==========================

try:
    from loguru import logger
    logger.remove()  # Remove default 'stderr' sink (and all others, if any)
    # logger.add('logs/autoassume_check.log', mode="w")
    logger.add("logs/statsassume_check.log", rotation="10 MB", mode="w")
except ImportError:
    import logging
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
    logger = logging.getLogger("statsassume")
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)



# ===========================
# Module: Statistical Tests
# Author: Kenneth Leung
# Last Modified: 12 Jan 2022
# ===========================
import pandas as pd
import statsmodels.stats.diagnostic as smd
import statsmodels.stats.stattools as sms
from statsmodels.stats.outliers_influence import variance_inflation_factor
import scipy.stats as stats
import warnings
from .utils import _convert_stat_table_to_dashtable

warnings.filterwarnings("ignore")


# ---------------------
#   Homoscedasticity
# ---------------------
def _get_interpretation_homosced(test_name: str,
                                 pvalue: float,
                                 sig_level: float):
    """Displays p-value of the specific homoscedasticity test, along with the test result interpretation

    Args:
        test_name (str): Name of statistical test
        pvalue (float): Value of the p-value
        sig_level (float): Significance level of statistical test

    Returns:
        str: Interpretation of homoscedasticity statistical test
    """

    if pvalue < sig_level:
        interpretation = f'p-value of {test_name} ({round(pvalue,3)}) is <{sig_level}, suggesting that the assumption of homoscedasticity is VIOLATED'
    else:
        interpretation = f'p-value of {test_name} ({round(pvalue,3)}) is ≥{sig_level}, suggesting that the assumption of homoscedasticity is satisfied'

    return interpretation


def stat_breuschpagan(residuals: pd.Series,
                      X_constant: pd.DataFrame,
                      sig_level: float = 0.05):
    """Run Breusch-Pagan test (for homoscedasticity check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model
        X_constant (pd.DataFrame): Dataframe with predictor variables, and with constant (intercept) included
        sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

    Returns:
        Interpretation (str) and results (HTML table) of Breusch-Pagan test
    """

    test_name = 'Breusch-Pagan Test'
    test_df = pd.DataFrame(smd.het_breuschpagan(residuals, X_constant),
                           columns=['Value'],
                           index=['Lagrange multiplier (LM) statistic',
                                  'LM p-value', 'F statistic', 'F p-value'])
    test_df['Value'] = test_df['Value'].round(decimals=3)
    pvalue = test_df.loc['LM p-value'].values[0]

    # Display results and interpretation
    interpretation = _get_interpretation_homosced(test_name, pvalue, sig_level)

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_df.columns = ['Parameter', 'Value']
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


def stat_white(residuals: pd.Series,
               X_constant: pd.DataFrame,
               sig_level: float = 0.05):
    """Run White test (for homoscedasticity check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model
        X_constant (pd.DataFrame): Predictor dataframe with constant (intercept) included
        sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

    Returns:
        Interpretation (str) and results (HTML table) of White test
    """

    test_name = 'White Test'
    test_df = pd.DataFrame(smd.het_white(residuals, X_constant),
                           columns=['Value'],
                           index=['Lagrange multiplier (LM) statistic',
                                  'LM p-value', 'F statistic', 'F p-value'])
    test_df['Value'] = test_df['Value'].round(decimals=3)
    pvalue = test_df.loc['LM p-value'].values[0]

    # Display results and interpretation
    interpretation = _get_interpretation_homosced(test_name, pvalue, sig_level)

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_df.columns = ['Parameter', 'Value']
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


def stat_gq(residuals: pd.Series,
            X_constant: pd.DataFrame,
            sig_level: float = 0.05):
    """Run Goldfeld-Quandt test (for homoscedasticity check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model
        X_constant (pd.DataFrame): Predictor dataframe with constant (intercept) included
        sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

    Returns:
        Interpretation (str) and results (HTML table) of Goldfeld-Quandt test
    """

    test_name = 'Goldfeld-Quandt Test'
    test_df = pd.DataFrame(smd.het_goldfeldquandt(residuals, X_constant)[:-1],
                           columns=['value'],
                           index=['F statistic', 'F p-value'])
    pvalue = test_df.loc['F p-value'].values[0]

    # Display results and interpretation
    interpretation = _get_interpretation_homosced(test_name, pvalue, sig_level)

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_df.columns = ['Parameter', 'Value']
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


# -----------------------------------
#     Independence / Autocorrelation
# -----------------------------------
def stat_durbin_watson(residuals: pd.Series):
    """Run Durbin-Watson test (for independence/autocorrelation check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model

    Returns:
        Interpretation (str) and results (HTML table) of Durbin-Watson test
    """

    test_name = 'Durbin-Watson Test'
    statistic = round(sms.durbin_watson(residuals), 3)
    lower_thresh, ideal, upper_thresh = 1.5, 2, 2.5
    test_df = pd.DataFrame(statistic,
                           columns=['Value'],
                           index=['Durbin-Watson Statistic'])

    # Display interpretation of test result
    if statistic < lower_thresh:
        interpretation = f'The {test_name} statistic of {statistic} is < {lower_thresh}, suggesting POSITIVE autocorrelation of residuals, and that assumption of observation independence is VIOLATED'
    elif statistic > upper_thresh:
        interpretation = f'The {test_name} statistic of {statistic} is > {upper_thresh}, suggesting NEGATIVE autocorrelation of residuals, and that assumption of observation independence is VIOLATED'
    else:
        interpretation = f'The {test_name} statistic of {statistic} is close to the value of {ideal}, suggesting NO autocorrelation of residuals, and that assumption of independence is satisfied'

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_df.columns = ['Parameter', 'Value']
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


def stat_ljungbox(residuals: pd.Series,
                  sig_level: float = 0.05,
                  lags: int = None,
                  auto_lag: bool = True):
    """Run Ljung-Box test (for independence/autocorrelation check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model
        sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.
        lags (int, optional): Number of lags to test. Defaults to None.
        auto_lag (bool, optional): Flag indicating whether to automatically determine
            optimal lag length based on threshold of maximum correlation value. Defaults to True

    Returns:
        Interpretation (str) and results (HTML table) of Ljung-Box test
    """

    test_name = 'Ljung-Box Test'
    test_df = smd.acorr_ljungbox(residuals, lags=lags, auto_lag=auto_lag)
    test_df.columns = ['Ljung-Box statistic', 'p-value']
    test_df = test_df.round(3)

    pvalues = test_df['p-value'].values
    min_pvalue = round(float(min(pvalues)), 3)

    if min_pvalue < sig_level:
        interpretation = f'The p-value of {test_name} ({min_pvalue}) is <{sig_level}, suggesting that the assumption of observation independence (aka no autocorrelation) is VIOLATED'
    else:
        interpretation = f'The p-value of {test_name} ({min_pvalue}) is ≥{sig_level}, suggesting that the assumption of observation independence (aka no autocorrelation) is satisfied'

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


# ---------------------
#   Multicollinearity
# ---------------------
def stat_vif(X_constant: pd.DataFrame,
             threshold: int = 10):
    """Calculate Variance Inflation Factor (VIF) values for each variable (for multi-collinearity check). This is done on X_constant,
    where the constant (intercept) is already included in the VIF calculations.

    Args:
        X_constant (pd.DataFrame): Dataframe with predictor variables, and with constant (intercept) included
        threshold (int, optional): Threshold for VIF to indicate multicollinearity. Defaults to 10.

    Returns:
        - pandas.DataFrame: Dataframe containing VIF value for each predictor variable
        - str: Interpretation of VIF values based on threshold
    """

    vif = [variance_inflation_factor(X_constant.values, i) for i in range(X_constant.shape[1])]
    X_cols = [col for col in list(X_constant.columns) if col != 'const']  # Hide const value from table, but VIF calc ALREADY correctly done on X_constant
    test_df = pd.DataFrame({'VIF': vif[1:]}, index=X_cols)
    test_df.sort_values(by='VIF', inplace=True, ascending=False)
    test_df['Below threshold'] = test_df['VIF'].apply(lambda x: u'\u2713' if x < threshold else 'X')
    test_df['VIF'] = test_df['VIF'].round(decimals=1)

    count_below_thresh = (test_df['Below threshold'].values != 'X').sum()
    count_above_thresh = len(test_df) - count_below_thresh
    if count_above_thresh > 0:
        interpretation = f'Given there are {count_above_thresh} features with VIF greater than threshold value of {threshold}, the assumption of no multicollinearity is VIOLATED'
    else:
        interpretation = f'Given zero features with VIF greater than threshold value of {threshold}, the assumption of no multicollinearity is satisfied'

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_df.columns = ['Feature', 'VIF', 'Below Threshold']
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


# ---------------------
#       Normality
# ---------------------
def _get_interpretation_normality(test_name: str,
                                  pvalue: float,
                                  sig_level: float):
    """Displays p-value of the specific normality test,
    along with test result interpretation

    Args:
        test_name (str): Name of statistical test
        pvalue (float): Value of the p-value
        sig_level (float): Significance level of statistical test

    Returns:
        str: Interpretation of normality statistical test
    """
    if pvalue < sig_level:
        interpretation = f'p-value of {test_name} ({round(pvalue,3)}) is <{sig_level}, suggesting the assumption of residual normality is VIOLATED'
    else:
        interpretation = f'p-value of {test_name} ({round(pvalue,3)}) is ≥{sig_level}, suggesting the assumption of residual normality is satisfied'

    return interpretation


def stat_anderson(residuals: pd.Series,
                  sig_level: float = 0.05):
    """Run Anderson-Darling test (for residual normality check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model
        sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

    Returns:
        Interpretation (str) and results (HTML table) of Anderson-Darling test
    """

    test_name = 'Anderson-Darling test'
    statistic = stats.anderson(residuals, dist='norm').statistic
    pvalue = round(stats.anderson(residuals, dist='norm').critical_values[2], 3)  # Get 5% critical value (significance level)

    test_df = pd.DataFrame.from_dict({'Anderson-Darling statistic': statistic,
                                     'p-value': pvalue},
                                     orient='index',
                                     columns=['Value'])

    # Display results and interpretation
    interpretation = _get_interpretation_normality(test_name, pvalue, sig_level)

    # Convert pandas df to dash table
    test_df.reset_index(drop=True, inplace=True)
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


def stat_shapiro(residuals: pd.Series,
                 sig_level: float = 0.05):
    """Run Shapiro-Wilk test (for residual normality check) and display test results

    Args:
        residuals (pd.Series): Residual values from regression model
        sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

    Returns:
        Interpretation (str) and results (HTML table) of Shapiro-Wilk test
    """

    test_name = 'Shapiro-Wilk test'
    statistic = stats.shapiro(residuals).statistic
    pvalue = round(stats.shapiro(residuals).pvalue, 3)

    test_df = pd.DataFrame.from_dict({'Shapiro-Wilk statistic': statistic,
                                     'p-value': pvalue},
                                     orient='index',
                                     columns=['Value'])
    test_df['Value'] = test_df['Value'].round(decimals=3)

    # Display results and interpretation
    interpretation = _get_interpretation_normality(test_name, pvalue, sig_level)

    # Convert pandas df to dash table
    test_df.reset_index(drop=False, inplace=True)
    test_df.columns = ['Parameter', 'Value']
    test_table = _convert_stat_table_to_dashtable(test_df)

    return interpretation, test_table


# def stat_lilliefors(residuals: pd.Series,
#                     sig_level: float = 0.05):
#     """Run Lilliefors corrected Kolmogorov-Smirnov test (for residual normality check) and display test results

#     Args:
#         residuals (pd.Series): Residual values from regression model
#         sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

#     Returns:
#         Interpretation (str) and results (HTML table) of Lilliefors corrected Kolmogorov-Smirnov test
#     """

#     test_name = 'Lilliefors corrected Kolmogorov-Smirnov test'
#     statistic = smd.lilliefors(residuals, 'norm', 'table')[0]
#     pvalue = smd.lilliefors(residuals, 'norm', 'table')[1]

#     # Display results and interpretation
#     interpretation = _get_interpretation_normality(test_name, pvalue, sig_level)

#     return interpretation


# def stat_jarque_bera(residuals: pd.Series,
#                      sig_level: float = 0.05):
#     """Run Jarque-Bera test (for residual normality check) and display test results

#     Args:
#         residuals (pd.Series): Residual values from regression model
#         sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

#     Returns:
#         Interpretation (str) and results (HTML table) of Jarque-Bera test
#     """

#     test_name = 'Jarque-Bera test'
#     statistic = stats.jarque_bera(residuals).statistic
#     pvalue = stats.jarque_bera(residuals).pvalue

#     # Display results and interpretation
#     interpretation = _get_interpretation_normality(test_name, pvalue, sig_level)

#     return interpretation


# def stat_ks(residuals: pd.Series,
#             sig_level: float = 0.05):
#     """Run Kolmogorov-Smirnov test (for residual normality check) and display test results

#     Args:
#         residuals (pd.Series): Residual values from regression model
#         sig_level (float, optional): Significance level of statistical test. Defaults to 0.05.

#     Returns:
#         Interpretation (str) and results (HTML table) of Kolmogorov-Smirnov test
#     """

#     test_name = 'Kolmogorov-Smirnov test'
#     statistic = stats.kstest(residuals, 'norm').statistic
#     pvalue = stats.kstest(residuals, 'norm').pvalue

#     # Display results and interpretation
#     interpretation = _get_interpretation_normality(test_name, pvalue, sig_level)

#     return interpretation



# ===========================================
# Module: Explanations of Assumption Checks
# Author: Kenneth Leung
# Last Modified: 22 Feb 2022
# ===========================================

# ---------------------------------
#   Overview - Linear Regression
# ----------------------------------
explain_description_linear_regression = """
There are a number of assumptions that justify the use of linear regression models for purposes \
of inference or prediction.

[1] Linearity of relationship between dependent and independent variables
[2] Statistical independence of errors (in particular, no correlation between consecutive \
errors (aka autocorrelation) for time series data)
[3] Homoscedasticity (constant variance) of errors
[4] Normality of error distribution
[5] No multicollinearity between independent variables

If any of the assumptions is violated, then the forecasts, confidence intervals, and scientific \
insights yielded by a regression model may be (at best) inefficient or (at worst) seriously biased \
or misleading.
"""

explain_solution_linear_regression = """
The best way to assess whether the dataset meets the assumptions of linear regression is to \
perform a series of assumption checks on the data.

StatsAssume simplifies the process of assumption checks by automatically running the statistical \
tests and generating visual plots for easy interpretation.

The results and information of each assumption check can be found in the Tabs above. Within each \
tab, there are also suggested strategies for handling the assumptions which have been violated.
"""

# ---------------------
#   Homoscedasticity
# ---------------------
explain_description_homosced = """
The variance of the residual errors should be constant across all values of the independent \
variables. Violations of homoscedasticity make it difficult to gauge the true standard \
deviation of the forecast errors, usually resulting in confidence intervals that are too \
wide or too narrow. This violation may also have the effect of giving too much weight to \
a small subset of the data (where error variance is largest) when estimating coefficients.
"""

explain_solution_homosced = """
If the dependent variable is strictly positive and if the residual-versus-predicted plot \
shows that the size of the errors is proportional to the size of the predictions (i.e., \
if errors seem consistent in percentage rather than absolute terms), a log transformation \
applied to the dependent variable may be appropriate. Heteroscedasticity can also be a \
byproduct of a significant violation of the linearity and/or independence assumptions, \
in which case it may also be fixed as a byproduct of fixing those problem.
"""

explain_plot_residual_homosced = """
The points and reference line in the residual plot should be horizontal, and should NOT be systematically getting larger or smaller \
in one direction by a significant amount (i.e., funnel shape).
"""

# ---------------------
#    Independence
# ---------------------
explain_description_independence = """
Residuals should be statistically independent of each other i.e., no autocorrelation \
(i.e., no correlation between consecutive errors or errors separated by some other \
number of periods). If in fact there is correlation among the error terms, then the \
estimated standard errors will tend to underestimate the true standard errors. As a \
result, confidence and prediction intervals will be narrower than they should be, and \
we may have an unwarranted sense of confidence in our model.

For example, a 95 percent confidence interval may in reality have a much lower \
probability than 0.95 of containing the true value of the parameter. In addition, \
pvalues will be lower than they should be, causing us to erroneously conclude that \
a parameter is statistically significant.

Autocorrelation is sometimes due to a violation of the linearity assumption. \
Autocorrelation also means that there is room for improvement in the model, \
and extreme autocorrelation is often a symptom of a badly mis-specified model.
"""

explain_solution_independence = """
For non-time-series models, resolve (if any) the violation of linearity assumption \
or bias that is explainable by omitted variables (e.g., interaction terms or dummies \
for identifiable conditions).

For time-series models, consider adding lags of the dependent variable and/or lags \
of some of the independent variables for minor cases of positive autocorrelation.
"""

explain_plot_acf = """
Most of the residual autocorrelation points should ideally fall randomly and \
symmetrically around zero and within the 95 percent (blue-coloured) confidence bands.
"""

# ---------------------
#      Linearity
# ---------------------
explain_description_linearity = """
There should be a linear relationship between dependent and independent variables. \
A linear relationship suggests that a change in Y due to one unit change in X \
is constant, regardless of the value of X.

If you fit a linear model to data which are nonlinearly or nonadditively related, \
your predictions are likely to be flawed, especially when you extrapolate beyond \
the range of the sample data.
"""

explain_solution_linearity = """
Consider applying a non-linear transformation to the dependent and/or independent variables \
as deemed appropriate and logical. For example, if the data are strictly positive, log \
transformation is an option. If a log transformation is applied to the dependent variable \
only, it assumes that it grows (or decays) exponentially as a function of the independent \
variables. If log transformation is applied to both dependent variable and independent \
variables, it assumes that the effects of independent variables are multiplicative rather \
than additive in their original units.

Another possibility is adding another regressor that is a nonlinear function of \
one of the other variables. If you have regressed Y on X, and the residual-vs-predicted \
plot shows a parabolic curve, then it may make sense to regress Y on both X and X^2.

Finally, it may be that you have overlooked some entirely different independent variable that \
explains the nonlinear pattern or interactions among variables seen in the residual plots. \
In that case, the shape of the pattern, together with economic or physical reasoning, may \
suggest some likely suspects. For example, if the strength of the linear relationship between Y \
and X1 depends on the level of some other variable X2, this could perhaps be addressed by creating \
a new independent variable that is the product of X1 and X2.
"""

explain_plot_residual_linearity = """
The points in the residual plot should be symmetrically distributed around the horizontal reference line \
with a roughly constant spread.
"""

explain_plot_pairplot = """
Each independent variable (x-axis) should have an approximate linear (straight-line) \
relationship with the dependent variable (y-axis).
"""

# ---------------------
#   Multicollinearity
# ---------------------
explain_description_multicollinearity = """
Independent variables should not be correlated with each other, meaning that \
there should be an absence of multicollinearity. When independent variables \
are correlated, it means that changes in one independent variable are associated \
with shifts in another independent variable. The stronger the correlation, the \
more difficult it becomes for the model to estimate the relationship between \
each independent variable and the dependent variable since the independent \
variables will tend to change in unison.

Multicollinearity reduces the precision of the estimated coefficients (and \
increases standard errors of estimates), which thus weakens the statistical \
power of your regression model. As such, it reduces the power of the model, \
and one may not be able to trust the p-values to identify independent variables \
that are statistically significant.
"""

explain_solution_multicollinearity = """
Some solutions include remove some of the highly correlated independent \
variables, linearly combining the independent variables (e.g.,adding them \
together) to form a new variable, or perform analyses designed for highly \
correlated variables like principal components analysis (PCA) or partial least \
squares regression.
"""

explain_plot_corrmatrix = """
An element within this heatmap matrix that is large in absolute value indicates \
a pair of highly correlated variables, and therefore suggests a collinearity \
issue in the data.
"""

# ---------------------
#      Normality
# ---------------------
explain_description_normality = """
The residuals must be approximately normally (aka Gaussian) distributed. Violations \
of normality create problems for determining whether model coefficients are significantly \
different from zero and for calculating confidence intervals for predictions. Since  \
parameter estimation is based on minimization of squared error, a few extreme observations \
can exert a disproportionate influence on parameter estimates. If the error distribution \
is significantly non-normal, confidence intervals may be too wide or too narrow.
"""

explain_solution_normality = """
Violations of normality often arise either because (a) distributions of dependent \
and/or independent variables are themselves significantly non-normal, and/or (b) \
the linearity assumption is violated. In such cases, a nonlinear transformation \
(e.g., log transformation) of variables might cure both problems.
"""

explain_plot_qq = """
If the distribution is normal, the points on the QQ plot should fall close to \
the diagonal reference line. An S-shaped pattern of deviations indicates that \
the residuals have excessive kurtosis.
"""

explainer_plot_residual_histogram = """
The histogram plot and kernel density estimation (KDE) curve should be approximately \
bell-shaped and symmetric about the mean.
"""
